# standard imports
import math
import os
import logging

# local imports
from .base import LevelDir

logg = logging.getLogger(__name__)


class NumDir(LevelDir):

    def __init__(self, root_path, thresholds=[1000]):
        thresholds = self.__thresholds_sanity(thresholds)
        super(NumDir, self).__init__(root_path, len(thresholds), 8)
        self.thresholds = thresholds
        fi = os.stat(self.master_file)
        self.entry_length = 8


    def __thresholds_sanity(self, thresholds):
        if len(thresholds) == 0:
            raise ValueError('thresholds must have at least one value')
        last_t = thresholds[0]
        for i in range(len(thresholds) - 1):
            if thresholds[i+1] > last_t:
                raise ValueError('thresholds must have diminishing order')
        return thresholds


    def to_dirpath(self, n): 
        c = n 
        x = 0
        d = []
        v = 0
        logg.debug('dirpath {}'.format(n))
        for t in self.thresholds:
            x = math.floor(c / t)
            y = x * t
            v += y
            d.append(str(v))
            c -= y
        return os.path.join(self.path, *d) 
      

    def to_filepath(self, n):
        path = self.to_dirpath(n)
        return os.path.join(path, str(n))


    def add(self, n, content, prefix=b''):
        entry_path = self.to_filepath(n)

        os.makedirs(os.path.dirname(entry_path), exist_ok=True)

        f = open(entry_path, 'wb')
        f.write(content)
        f.close()

        f = open(self.master_file, 'ab')
        f.write(n.to_bytes(8, byteorder='big'))
        f.close()

        c = self.count()

        logg.debug('created new numdir entry {} idx {} in {}'.format(str(n), c, entry_path)) 



# standard imports
import os
import stat


class LevelDir:

    def __init__(self, root_path, levels, entry_length):
        self.path = root_path
        self.levels = levels 
        self.entry_length = entry_length
        fi = None
        self.__prepare_directory(self.path)

        self.__verify_directory()

        self.master_file = os.path.join(self.path, '.master')


    def have(self, k):
        fp = self.to_filepath(k)
        try:
            os.stat(fp)
        except FileNotFoundError:
            return False
        return True


    def __verify_directory(self):
        fi = os.stat(self.path)
        if not stat.S_ISDIR(fi.st_mode):
            raise ValueError('{} is not a directory'.format(self.path))
        #f = os.listdir(self.path)
        #os.listdir(self.path)
        #f.close()
        return True


    def count(self):
        fi = os.stat(self.master_file)
        c = fi.st_size / self.entry_length
        r = int(c)
        if r != c: # TODO: verify valid for check if evenly divided
            raise IndexError('master file not aligned')
        return r


    @classmethod
    def __prepare_directory(cls, path):
        os.makedirs(path, exist_ok=True)
        state_file = os.path.join(path, '.master')
        try:
            os.stat(state_file)
        except FileNotFoundError:
            f = open(state_file, 'w')
            f.close()
        return state_file



# standard imports
import os
import logging

# external imports
from hexathon import valid as valid_hex

# local imports
from .base import LevelDir

logg = logging.getLogger(__name__)


def default_formatter(hx):
    return hx.upper()


class HexDir(LevelDir):

    def __init__(self, root_path, key_length, levels=2, prefix_length=0, formatter=default_formatter):
        super(HexDir, self).__init__(root_path, levels, key_length + prefix_length)
        #self.path = root_path
        self.key_length = key_length
        self.prefix_length = prefix_length
        self.__levels = levels + 2
        self.formatter = formatter


    def __check(self, key, content, prefix):
        l = len(key)
        if l != self.key_length:
            raise ValueError('expected key length {}, got {}'.format(self.key_length, l))
        l = len(prefix)
        if l != self.prefix_length:
            raise ValueError('expected prefix length {}, got {}'.format(self.prefix_length, l))
        if not isinstance(content, bytes):
            raise ValueError('content must be bytes, got {}'.format(type(content).__name__))
        if prefix != None and not isinstance(prefix, bytes):
            raise ValueError('prefix must be bytes, got {}'.format(type(content).__name__))


    def add(self, key, content, prefix=b''):
        self.__check(key, content, prefix)
        key_hex = self.key_to_string(key)
        entry_path = self.to_filepath(key_hex)
        return self.__add(entry_path, key, content, key_hex, prefix=prefix)


    def __add(self, entry_path, key, content, display_key, prefix=b''):
        c = self.count()

        os.makedirs(os.path.dirname(entry_path), exist_ok=True)
        f = open(entry_path, 'wb')
        f.write(content)
        f.close()

        f = open(self.master_file, 'ab')
        if prefix != None:
            f.write(prefix)
        f.write(key)
        f.close()

        logg.debug('created new hexdir entry {} idx {} in {}'.format(display_key, c, entry_path)) 
    
        return (c, entry_path)


    def key_to_string(self, k):
        return k.hex() 
       
    
    def add_dir(self, file_key, key, content, prefix=b''):
        self.__check(key, content, prefix)
        key_hex = self.key_to_string(key)
        entry_path = self.to_filepath(key_hex)
        entry_path = os.path.join(entry_path, file_key)
        return self.__add(entry_path, key, content, key_hex, prefix=prefix)


    def __cursor(self, idx):
        return idx * (self.prefix_length + self.key_length)


    def set_prefix(self, idx, prefix):
        l = len(prefix)
        if l != self.prefix_length:
            raise ValueError('expected prefix length {}, got {}'.format(self.prefix_length, l))
        if not isinstance(prefix, bytes):
            raise ValueError('prefix must be bytes, got {}'.format(type(content).__name__))
        cursor = self.__cursor(idx)
        f = open(self.master_file, 'rb+')
        f.seek(cursor)
        f.write(prefix)
        f.close()


    def get(self, idx):
        cursor = self.__cursor(idx)
        print('cursor {}'.format(cursor))
        f = open(self.master_file, 'rb')
        f.seek(cursor)
        prefix = f.read(self.prefix_length)
        key = f.read(self.key_length)
        f.close()
        return (prefix, key)


    def to_subpath(self, hx):
        lead = ''
        for i in range(0, self.__levels, 2):
            lead += hx[i:i+2] + '/'
        return self.formatter(lead)


    def to_dirpath(self, hx):
        sub_path = self.to_subpath(hx)
        return os.path.join(self.path, sub_path)


    def to_filepath(self, hx):
        dir_path = self.to_dirpath(hx)
        file_path = os.path.join(dir_path, self.formatter(hx))
        return file_path



from setuptools import setup
import os


requirements = []
f = open('requirements.txt', 'r')
while True:
    l = f.readline()
    if l == '':
        break
    requirements.append(l.rstrip())
f.close()


setup(
        install_requires=requirements,
    )

